import bz2
import numpy
import sys

from scipy.optimize import leastsq

try:
    from largescale.settings import INFTY, CALIBRATION
except ImportError:
    INFTY=numpy.inf
    CALIBRATION=163.6723370552

def compute_scores(out, labels):
    auPRC=None
    aoPRC=None
    auROC=None
    aoROC=None
    CC=None
    F1=None
    WRAcc=None
    acc=None
    err=None
    BAL=None

    auPRC = calcprcscore(out, labels)
    aoPRC = 1-auPRC
    auROC = calcrocscore(out, labels)
    aoROC = 1-auROC
    CC = calccc(out, labels)
    F1 = calcf1(out, labels)
    WRAcc = calcwracc(out, labels)
    acc = accuracy(out, labels)
    err = error(out, labels)
    BAL = calcbal(out, labels)

    return { 'aoPRC': aoPRC, 'auPRC': auPRC, 'aoROC': aoROC, 
            'auROC': auROC, 'CC': CC, 'F1': F1, 'WRAcc': WRAcc,
            'Accuracy': acc, 'Error': err, 'BAL': BAL }

def read_submission(outf, labf, dataset, track='SVM', output_name=None, check_track=False):
    """
        dataset_size0 -1 traintime0 testtime0 calibration0 objective C epsilon rfb-sigma output0 output1 ...
        ...
        dataset_sizeN -1 traintimeN testtimeN calibrationN objective C epsilon rfb-sigma output0 output1 ...
        dataset_sizeN index0 traintime0 testtime0 calibration0 objective C epsilon rfb-sigma output0 output1 ...
        ...
        dataset_sizeN index9 traintime9 testtime0 calibration9 objective C epsilon rfb-sigma output0 output1 ...
>a singly empty line here distinguishes wild from model specific track
        dataset_size0 -1 traintime0 testtime0 calibration0 objective C epsilon rfb-sigma
        ...
        dataset_sizeN -1 traintimeN testtimeN calibrationN objective C epsilon rfb-sigma
        dataset_sizeN index1 traintime1 testtime1 calibration1 objective C epsilon rfb-sigma
        ...
        dataset_sizeN index9 traintime9 testtime0 calibration9 objective C epsilon rfb-sigma
        dataset_sizeN -2 traintime0 testtime0 calibration0 objective C1 epsilon rfb-sigma
        ...
        dataset_sizeN -2 traintimeK testtimeK calibration9 objective CK epsilon rfb-sigma

(or -3 and rbf-sigma 1 ... K)

        Explanation of values
        =====================
        * dataset size values must match size of dataset or 10^[2,3,4,5,6,7]
        * index is 0...9 (to distinguish values obtained while optimizing, -1 is used)
        * traintime is the measured traintime in seconds
        * testtime is the measured time needed for testing in seconds
        * calibration is the score obtained using the calibration script
        (values should be the same if run on the same machine)

        * Method specific values, objective, C, rbf-sigma, epsilon
          - use 0 if not applicable
          - for SVM please fill in the different parameters, i.e. SVM objective,
          SVM-C, rbf-sigma and stopping condition epsilon, as the meaning of epsilon may differ
          please explain in the description what epsilon stands for
      - for the linear svm track index -2 should be used when doing the experiment for C
      - for the gaussian svm track index -3 should be used when modifying rbf sigma

        * the number of values must match the size of the validation/test set
    """
    line=0

    wild_outputs=list()
    more_outputs=None
    lab=numpy.array([ float(x) for x in labf.readlines()], dtype=numpy.float32)
    datasetsizes={'dna': 50000000,
            'webspam' : 350000,
            'face' : 5469800,
            'ocr' : 3500000,
            'alpha' : 500000,
            'beta' : 500000,
            'gamma' : 500000,
            'delta' : 500000,
            'epsilon' : 500000,
            'zeta' : 500000 }
    true_datasize=datasetsizes[dataset]

    for l in outf.readlines():
        line+=1
        l=l.strip()

        if len(l)<2:
            if check_track and track=='Wild':
                raise SystemError('Parallel or SVM track outputs found but not submitting to parallel/SVM track')
            if more_outputs is not None:
                raise SystemError("No further empty lines allowed after starting parallel or SVM track section (check for trailing empty lines)")
            else:
                more_outputs=list()
            continue

        try:
            out=[ float(x) for x in l.split()]
        except:
            raise SystemError('Non floating point number found in line %d.' % line)

        if len(out)!=len(lab)+9:
            if track in ('SVM', 'Parallel'):
                if len(out)!=9:
                    raise SystemError('Length of output in line %d does not match number '
                            'of labels+9 (%d vs %d) nor 9' % (line, len(out), len(lab)+9))
            else:
                raise SystemError('Length of output in line %d does not match number ' 
                        'of labels+9 (%d vs %d)' % (line, len(out), len(lab)+9))

        vals=numpy.array(out[:9], dtype=numpy.float)
        if len(out)>9:
            scores=compute_scores(numpy.array(out[9:], dtype=numpy.float), lab)
        else:
            scores=None

        out=(vals,scores)

        if more_outputs != None:
            more_outputs.append(out)
        else:
            wild_outputs.append(out)

    if check_track:
        if more_outputs is None and track=='SVM':
            raise SystemError('No outputs for C/epsilon for SVMs found but submitting to SVM track')
        if more_outputs is None and track=='Parallel':
            raise SystemError('No outputs for parallel track found but submitting to parallel track')
        if more_outputs and track=='Wild':
            raise SystemError('Parallel or SVM track outputs found but not submitting to parallel/SVM track')
        if len(wild_outputs)==0:
            raise SystemError('No submissions found')

    return compute_compo_scores_and_figures(wild_outputs, more_outputs, track, true_datasize,
            output_name!="final", perfname='aoPRC')

#svm
def compute_compo_score_svmtimevsobjective(scores, c, ub_obj=0.05):
    scores['svm_area_timevsobj']=compute_area(c[0], c[1])
    scores['svm_min_obj']=numpy.min(c[1])
    t=INFTY
    i=numpy.where(((c[1]-ub_obj)/c[1])<0.05)
    if len(i[0])>0:
        t=numpy.min(c[0][i])

    scores['svm_time_for_fixed_obj']=t

def compute_compo_score_svmcvstime(scores, c):
    scores['svm_avg_cvstime']=numpy.mean(c[1])

def compute_compo_score_svmsigmavstime(scores, s):
    scores['svm_avg_sigmavstime']=numpy.mean(s[1])

def compute_compo_score_svmsizevstime(scores, c):
    scores['svm_effort_slope']=compute_slope(c[0], c[1])

#wild
def compute_compo_score_timevserror(scores, c, ub_err=0.05, infinity=INFTY):
    scores['area_timevserr']=compute_area(c[0], c[1])
    scores['min_error']=numpy.min(c[1])
    t=infinity
    i=numpy.where(((c[1]-ub_err)/c[1])<0.05)
    if len(i[0])>0:
        t=numpy.min(c[0][i])

    scores['time_for_fixed_error']=t

def compute_compo_score_sizevstime(scores, c):
    scores['effort_slope']=compute_slope(c[0], c[1])

def compute_compo_score_sizevstesttime(scores, c):
    scores['avg_testtime']=numpy.mean(c[1])

def sanitize_curve(x,y):
    i=numpy.argsort(x)
    return (x[i],y[i])

def compute_compo_score_sizevserror(scores, c, ub_err=0.05, infinity=INFTY):
    scores['area_sizevserr']=compute_area(c[0], c[1])

    t=infinity
    i=numpy.where(((c[1]-ub_err)/c[1])<0.05)
    if len(i[0])>0:
        t=numpy.min(c[0][i])

    scores['size_for_fixed_error']=t

def compute_compo_scores_and_figures(output, otheroutput, track, true_datasize, use_calibration, perfname):
    #WILD COMPETITION
    if output and len(output)>0:
        error=numpy.zeros(len(output)) #perf measure

        datasize=numpy.zeros(len(output)) #dataset size
        epoch=numpy.zeros(len(output)) #epoch measure
        traintime=numpy.zeros(len(output)) #traintime
        testtime=numpy.zeros(len(output)) #testtime
        loadingtime=numpy.zeros(len(output)) #loadingtime

        curves=dict()
        scores=dict()

        if track=='Parallel':
            for i in xrange(len(output)):
                e=output[i]
                datasize[i]=e[0][0]
                epoch[i]=e[0][1]
                traintime[i]=(e[0][2]-e[0][3])
                if use_calibration:
                    if e[0][4]<=0:
                        raise SystemError("calibration time <= 0")
                    traintime[i]=(CALIBRATION*traintime[i])/e[0][4] #(walltime-dataloadingtime)/calibr
                if traintime[i]<=0:
                    traintime[i]=1e-6;
                loadingtime[i]=e[0][3] #data loading time
                if loadingtime[i]<=0:
                    loadingtime[i]=1e-6;
                error[i]=e[1][perfname]
        else:
            for i in xrange(len(output)):
                e=output[i]
                datasize[i]=e[0][0]
                epoch[i]=e[0][1]
                traintime[i]=e[0][2] #traintime/calibr
                if use_calibration:
                    if e[0][4]<=0:
                        raise SystemError("calibration time <= 0")
                    traintime[i]=(CALIBRATION*traintime[i])/e[0][4]
                if traintime[i]<=0:
                    traintime[i]=1e-6;
                testtime[i]=e[0][3] #testtime/calibr
                if use_calibration:
                    testtime[i]=(CALIBRATION*testtime[i])/e[0][4]
                if testtime[i]<=0:
                    testtime[i]=1e-6;
                error[i]=e[1][perfname]

        #timevserror
        m=numpy.max(datasize)
        i=numpy.where((datasize==m) & (epoch!=-1))
        if not i or len(i[0])==0:
            raise SystemError("Missing intermediate progress experiments (no epochs>=0 for largest dataset found")
        c=sanitize_curve(traintime[i],error[i])
        if c[0][-1]<c[0][0]*10:
            curves['timevserror']=(numpy.concatenate((c[0],[c[0][0]*10])),
                    numpy.concatenate((c[1],[c[1][-1]])))
        else:
            curves['timevserror']=c
        compute_compo_score_timevserror(scores, curves['timevserror'])

        #sizevstime
        i=numpy.where(epoch==-1)
        if not i or len(i[0])==0:
            raise SystemError("Missing different dataset sizes experiment (no epochs=-1 for different dataset sizes found")
        curves['sizevstime']=sanitize_curve(datasize[i],traintime[i])
        compute_compo_score_sizevstime(scores, curves['sizevstime'])

        if track=='Parallel':
            curves['sizevsloadingtime']=sanitize_curve(datasize[i],testtime[i])
            compute_compo_score_sizevstesttime(scores, curves['sizevsloadingtime'])
        else:
            curves['sizevstesttime']=sanitize_curve(datasize[i],testtime[i])
            compute_compo_score_sizevstesttime(scores, curves['sizevstesttime'])

        #sizevserror
        i=numpy.where(epoch==-1)
        m=numpy.max(datasize[i])
        c=sanitize_curve(datasize[i],error[i])
        if m<true_datasize:
            curves['sizevserror']=(numpy.concatenate((c[0],[true_datasize])),
                    numpy.concatenate((c[1],[c[1][-1]])))
        else:
            curves['sizevserror']=c
        compute_compo_score_sizevserror(scores, curves['sizevserror'])
    else:
        raise SystemError("no outputs found, file is empty or contains trailing newline.")

    # SVMs or Parallel
    if otheroutput and len(otheroutput)>0:
        datasize=numpy.zeros(len(otheroutput)) #dataset size
        epoch=numpy.zeros(len(otheroutput)) #epoch measure
        traintime=numpy.zeros(len(otheroutput)) #traintime
        obj=numpy.zeros(len(otheroutput)) #SVM objective
        C=numpy.zeros(len(otheroutput)) #SVM C
        eps=numpy.zeros(len(otheroutput)) #epsilon
        rbf_sigma=numpy.zeros(len(otheroutput)) #rbf sigma

        if track=='Parallel':
            for i in xrange(len(otheroutput)):
                e=otheroutput[i]
                datasize[i]=e[0][0]
                epoch[i]=e[0][1]
                traintime[i]=(e[0][2]-e[0][3]) #(walltime-dataloadingtime)/calibr
                if use_calibration:
                    if e[0][4]<=0:
                        raise SystemError("calibration time <= 0")
                    traintime[i]=(CALIBRATION*traintime[i])/e[0][4]
                if traintime[i]<=0:
                    traintime[i]=1e-6;
                C[i]=e[0][6]

            # number of cpus vs time
            m=numpy.max(datasize)
            i=numpy.where((datasize==m) & (epoch==-4))
            if not i or len(i[0])==0:
                raise SystemError("Missing different training times for 1,2,4,8 CPUs (no epochs=-4 for maximum dataset size found")
            curves['cpuvstime']=sanitize_curve(traintime[i],C[i])
            compute_compo_score_svmcvstime(scores, curves['cpuvstime'])

        elif track=='SVM':
            for i in xrange(len(otheroutput)):
                e=otheroutput[i]
                datasize[i]=e[0][0]
                epoch[i]=e[0][1]
                traintime[i]=e[0][2] #traintime/calibr
                if use_calibration:
                    if e[0][4]<=0:
                        raise SystemError("calibration time <= 0")
                    traintime[i]=(CALIBRATION*traintime[i])/e[0][4]
                if traintime[i]<=0:
                    traintime[i]=1e-6;
                obj[i]=e[0][5]
                C[i]=e[0][6]
                eps[i]=e[0][7]
                rbf_sigma[i]=e[0][8]

            #timevsobjective
            m=numpy.max(datasize)
            i=numpy.where((datasize==m) & (epoch>=0))
            if not i or len(i[0])==0:
                raise SystemError("Missing different dataset sizes SVM experiment (no epochs=-1 for different dataset sizes found")
            curves['svmtimevsobjective']=sanitize_curve(traintime[i],obj[i])
            compute_compo_score_svmtimevsobjective(scores, curves['svmtimevsobjective'])

            #sizevstime
            i=numpy.where(epoch==-1)
            if not i or len(i[0])==0:
                raise SystemError("Missing different dataset sizes SVM experiment (no epochs=-1 for different dataset sizes found")
            curves['svmsizevstime']=sanitize_curve(datasize[i],traintime[i])
            compute_compo_score_svmsizevstime(scores, curves['svmsizevstime'])

            if not numpy.any(epoch==-2) and not numpy.any(epoch==-3):
                raise SystemError("Missing different C/sigma SVM experiment (no epochs=-2 or epochs=-3 for different C/sigma found")

            #cvstime
            i=numpy.where((datasize==m) & (epoch==-2))
            if i and len(i[0])>0:
                curves['svmcvstime']=sanitize_curve(C[i],traintime[i])
                compute_compo_score_svmcvstime(scores, curves['svmcvstime'])
            #raise SystemError("Missing different C SVM experiment (no epochs=-2 for different C found")

            #sigmavstime
            i=numpy.where((datasize==m) & (epoch==-3))
            if i and len(i[0])>0:
                curves['svmsigmavstime']=sanitize_curve(rbf_sigma[i],traintime[i])
                compute_compo_score_svmsigmavstime(scores, curves['svmsigmavstime'])
            #raise SystemError("Missing different sigma SVM experiment (no epochs=-3 for different sigma found")

    return  (curves, scores)

def calcbal(out, lab):
    pidx=(lab==+1)
    perr=numpy.mean(out[pidx]>0)
    nidx=(lab==-1)
    nerr=numpy.mean(out[nidx]<0)
    return 0.5 * (perr+nerr)

def accuracy(out, lab):
    return numpy.mean(numpy.sign(out) == lab)

def error(out, lab):
    return numpy.mean(numpy.sign(out) != lab)

def get_tp(out, lab):
    pidx=(lab==+1)
    return numpy.sum(numpy.sign(out[pidx])>0)

def get_tn(out, lab):
    nidx=(lab==-1)
    return numpy.sum(numpy.sign(out[nidx])<0)

def get_fp(out, lab):
    nidx=(lab==-1)
    return numpy.sum(numpy.sign(out[nidx])>0)

def get_fn(out, lab):
    pidx=(lab==+1)
    return numpy.sum(numpy.sign(out[pidx])<0)

def calcwracc(out, lab):
    tp=get_tp(out, lab)
    fp=get_fp(out, lab)
    tn=get_tn(out, lab)
    fn=get_fn(out, lab)
    return tp/(tp+fn) - fp/(fp+tn)

def calcf1(out, lab):
    tp=get_tp(out, lab)
    fp=get_fp(out, lab)
    fn=get_fn(out, lab)
    return 2.0*tp/(2.0*tp+fp+fn)

def calccc(out, lab):
    tp=float(get_tp(out, lab))
    fp=float(get_fp(out, lab))
    tn=float(get_tn(out, lab))
    fn=float(get_fn(out, lab))
    return (tp*tn-fp*fn)/numpy.sqrt((tp+fp)*(tp+fn)*(tn+fp)*(tn+fn))

def calcroc(out, lab):
    idx = numpy.argsort(out)
    tp=numpy.concatenate( ([1], 1-numpy.cumsum(lab[idx]>0)/float(numpy.sum(lab > 0))) )
    fp=numpy.concatenate( ([1], 1-numpy.cumsum(lab[idx]<0)/float(numpy.sum(lab < 0))) )
    return (fp,tp)

def calcrocscore(out, lab):
    a,b=calcroc(out,lab)
    score = 0.5*numpy.sum((a[:-1]-a[1:])*(b[:-1]+b[1:]))
    return score

def calcprc(out, lab):
    pmap=(lab==1)
    idx = numpy.argsort(-out)
    pmap=pmap[idx]

    recall=numpy.cumsum(pmap)/numpy.double(sum(lab==1))
    precision=numpy.cumsum(pmap)/(numpy.arange(len(pmap),dtype=numpy.double)+1.0)
    return (precision,recall)

def calcprcscore(out, lab):
    a,b=calcprc(out,lab)
    score = 0.5*numpy.sum((a[:-1]+a[1:])*(b[1:]-b[:-1]))
    return score

def compute_area(a,b):
    return 0.5*numpy.sum(numpy.abs(a[:-1]-a[1:])*numpy.abs(b[:-1]+b[1:]))

def poly_eval(x, p):
    a,b = p  
    return a*x**b

def residuals(p, y, x):
    return y-poly_eval(x,p)

def compute_slope(x,y):
    p0=[1.0,1.0]
    plsq = leastsq(residuals, p0, args=(y, x))
    return plsq[0][1]

def alternative_calcprcscore(out, lab):
    """
    code translated from image challenge
    """
    idx = numpy.argsort(-out)
    tp=lab[idx]>0
    fp=lab[idx]<0
    fp=1.0*numpy.cumsum(fp)
    tp=1.0*numpy.cumsum(tp)
    rec=tp/numpy.sum(lab>0)
    prec=tp/(fp+tp)
    ap=0
    for t in numpy.arange(0,1.01, 0.01):
        el=prec[rec>=t]
        if len(el)>0:
            p=numpy.max(el)
        else:
            p=0
        ap=ap+p/101.0
    return ap

def usage():
    print """%s 'output.bz2 label.bz2\n\n
    compute performance curve/measures for outputs\n

    auROC - area under Receiver Operator Characteristic Curve
    auPRC - area under Precision Recall Curve
    CC - cross correlation coefficient
    F1 - F1 score, ACC,BAL\n\n""" % sys.argv[0]

if __name__ == '__main__':
    if len(sys.argv) != 3:
        usage()
    else:
        outfname=sys.argv[1]
        labfname=sys.argv[2]
        if outfname.endswith('.bz2'):
            outf=bz2.BZ2File(outfname)
        else:
            outf=file(outfname)

        if labfname.endswith('.bz2'):
            labf=bz2.BZ2File(labfname)
        else:
            labf=file(labfname)

        (curves, scores)=read_submission(outf, labf)

        print "\nscores:\n======="
        for i in scores:
            print i,scores[i]

        try:
            from pylab import *

            figure()
            x,y=curves['timevserror']
            semilogx(x,y, marker='.', markersize=20, linewidth=3)
            xlabel('CPU Time (s)')
            ylabel('aoPRC')
            grid()

            figure()
            x,y=curves['sizevstime']
            loglog(x,y, marker='.', markersize=20, linewidth=3)
            xlabel('Dataset Size')
            ylabel('CPU Time (s)')
            grid()

            figure()
            x,y=curves['sizevserror']
            semilogx(x,y, marker='.', markersize=20, linewidth=3)
            xlabel('Dataset Size')
            ylabel('aoPRC')
            grid()

            if curves.has_key('svmtimevsobjective'):
                figure()
                x,y=curves['svmtimevsobjective']
                semilogx(x,y, marker='.', markersize=20, linewidth=3)
                xlabel('CPU Time (s)')
                ylabel('Objective')
                grid()

            if curves.has_key('svmsizevstime'):
                figure()
                x,y=curves['svmsizevstime']
                loglog(x,y, marker='.', markersize=20, linewidth=3)
                xlabel('Dataset Size')
                ylabel('CPU Time (s)')
                grid()

            if curves.has_key('svmcvstime'):
                figure()
                x,y=curves['svmcvstime']
                semilogx(x,y, marker='.', markersize=20, linewidth=3)
                xlabel('C')
                ylabel('Time')
                grid()

            #if curves.has_key('svmsigmavstime'):
            #    figure()
            #    x,y=curves['svmsigmavstime']
            #    semilogx(x,y, marker='.', markersize=20, linewidth=3)
            #    xlabel('Sigma')
            #    ylabel('Time')
            #    grid()
            show()
        except ImportError:
            print "Install matplotlib to get the figures"
